@@ -1,12 +1,22 @@
+"""Module for evaluating machine learning models using cross-validation and A/B testing.
+
+Provides functionality for model evaluation, comparison, and statistical testing.
+"""
+
 import logging
 from typing import Dict, List, Optional, Union
 
 import numpy as np
 import pandas as pd
 from scipy.stats import ttest_rel
-from sklearn.base import ClassifierMixin, RegressorMixin
+from sklearn.base import ClassifierMixin, RegressorMixin, is_classifier, is_regressor
 from sklearn.metrics import classification_report, confusion_matrix
-from sklearn.model_selection import cross_validate
+from sklearn.model_selection import (
+    KFold,
+    StratifiedKFold,
+    cross_val_predict,
+    cross_validate,
+)
 
 logging.basicConfig(level=logging.INFO, format="%(asctime)s %(levelname)s %(message)s")
 logger = logging.getLogger(__name__)
@@ -27,7 +37,7 @@ class ModelEvaluator:
         self,
         X: pd.DataFrame,
         y: pd.Series,
-        cv: int = 5,
+        cv: Union[int, StratifiedKFold, KFold] = 5,
         scoring: Optional[Union[str, List[str]]] = None,
     ) -> Dict[str, List[float]]:
         """Perform cross-validation and return scores.
@@ -35,7 +45,7 @@ class ModelEvaluator:
         Args:
             X: Features.
             y: Labels.
-            cv: Number of cross-validation folds.
+            cv: Number of cross-validation folds or a splitter object.
             scoring: Metric(s) to evaluate (default: accuracy for classifiers, r2 for regressors).
 
         Returns:
@@ -49,25 +59,27 @@ class ModelEvaluator:
             raise ValueError("Input data (X or y) is empty.")
 
         if scoring is None:
-            if isinstance(self.model, ClassifierMixin):
-                scoring = "accuracy"
-            elif isinstance(self.model, RegressorMixin):
-                scoring = "r2"
+            if is_classifier(self.model):
+                scoring = ["accuracy"]
+            elif is_regressor(self.model):
+                scoring = ["r2"]
             else:
                 raise ValueError("Model must be a classifier or regressor.")
+        elif isinstance(scoring, str):
+            scoring = [scoring]
 
         try:
             scores = cross_validate(self.model, X, y, cv=cv, scoring=scoring)
             return {metric: scores[f"test_{metric}"].tolist() for metric in scoring}
         except Exception as e:
             logger.error(f"Error occurred during cross-validation: {e}")
-            raise RuntimeError(f"Cross-validation failed: {e}")
+            raise RuntimeError(f"Cross-validation failed: {e}") from e
 
     def cross_validation_report(
         self,
         X: pd.DataFrame,
         y: pd.Series,
-        cv: int = 5,
+        cv: Union[int, StratifiedKFold, KFold] = 5,
         target_names: Optional[List[str]] = None,
     ) -> str:
         """Generate a cross-validation report.
@@ -75,53 +87,36 @@ class ModelEvaluator:
         Args:
             X: Features.
             y: Labels.
-            cv: Number of cross-validation folds.
+            cv: Number of cross-validation folds or a splitter object.
             target_names: Names of target classes (for classification).
 
         Returns:
             A string containing the cross-validation report.
 
         Raises:
-            ValueError: If the input data is invalid.
-            RuntimeError: If the model is not a classifier.
+            ValueError: If the input data is invalid or model is not a classifier.
+            RuntimeError: If report generation fails.
         """
         if X.empty or y.empty:
             raise ValueError("Input data (X or y) is empty.")
 
-        if not isinstance(self.model, ClassifierMixin):
+        if not is_classifier(self.model):
             raise ValueError("Model must be a classifier for classification metrics.")
 
         try:
-            cv_results = cross_validate(
-                self.model,
-                X,
-                y,
-                cv=cv,
-                return_train_score=False,
-                scoring=["accuracy", "precision", "recall", "f1"],
-            )
-
-            report = "Cross-Validation Metrics:\n"
-            for metric in ["accuracy", "precision", "recall", "f1"]:
-                mean_score = np.mean(cv_results[f"test_{metric}"])
-                report += f"{metric.capitalize()}: {mean_score:.4f}\n"
-
-            report += "\nClassification Report:\n"
-            report += classification_report(
-                y, self.model.predict(X), target_names=target_names
-            )
-
+            y_pred = cross_val_predict(self.model, X, y, cv=cv)
+            report = "Cross-Validation Classification Report:\n"
+            report += classification_report(y, y_pred, target_names=target_names)
             return report
         except Exception as e:
-            logger.error(
-                f"Error occurred while generating cross-validation report: {e}"
-            )
-            raise RuntimeError(f"Cross-validation report generation failed: {e}")
+            logger.error(f"Error generating cross-validation report: {e}")
+            raise RuntimeError("Cross-validation report failed.") from e
 
     def confusion_matrix_report(
         self,
         X: pd.DataFrame,
         y: pd.Series,
+        cv: Union[int, StratifiedKFold, KFold] = 5,
         labels: Optional[List[int]] = None,
     ) -> np.ndarray:
         """Generate a confusion matrix report.
@@ -129,29 +124,29 @@ class ModelEvaluator:
         Args:
             X: Features.
             y: Labels.
+            cv: Number of cross-validation folds or a splitter object.
             labels: List of labels to include in the confusion matrix.
 
         Returns:
             The confusion matrix.
 
         Raises:
-            ValueError: If the input data is invalid.
-            RuntimeError: If the model is not a classifier.
+            ValueError: If the input data is invalid or model is not a classifier.
+            RuntimeError: If report generation fails.
         """
         if X.empty or y.empty:
             raise ValueError("Input data (X or y) is empty.")
 
-        if not isinstance(self.model, ClassifierMixin):
+        if not is_classifier(self.model):
             raise ValueError("Model must be a classifier for classification metrics.")
 
         try:
-            cm = confusion_matrix(y, self.model.predict(X), labels=labels)
+            y_pred = cross_val_predict(self.model, X, y, cv=cv)
+            cm = confusion_matrix(y, y_pred, labels=labels)
             return cm
         except Exception as e:
-            logger.error(
-                f"Error occurred while generating confusion matrix report: {e}"
-            )
-            raise RuntimeError(f"Confusion matrix report generation failed: {e}")
+            logger.error(f"Error generating confusion matrix: {e}")
+            raise RuntimeError("Confusion matrix report failed.") from e
 
 
 def perform_ab_test(
@@ -159,7 +154,7 @@ def perform_ab_test(
     model_b: Union[ClassifierMixin, RegressorMixin],
     X: pd.DataFrame,
     y: pd.Series,
-    cv: int = 5,
+    cv: Union[int, StratifiedKFold, KFold] = 5,
     scoring: Optional[Union[str, List[str]]] = None,
 ) -> Dict[str, Dict[str, Union[float, List[float]]]]:
     """Perform A/B testing between two models.
@@ -169,28 +164,46 @@ def perform_ab_test(
         model_b: The second model to compare.
         X: Features.
         y: Labels.
-        cv: Number of cross-validation folds.
+        cv: Number of cross-validation folds or a splitter object.
         scoring: Metric(s) to evaluate (default: accuracy for classifiers, r2 for regressors).
 
     Returns:
         Dictionary containing A/B test results.
 
     Raises:
-        ValueError: If the input data is invalid.
+        ValueError: If input data is invalid or models are not the same type.
         RuntimeError: If A/B testing fails.
     """
     if X.empty or y.empty:
         raise ValueError("Input data (X or y) is empty.")
 
+    if not (
+        (is_classifier(model_a) and is_classifier(model_b))
+        or (is_regressor(model_a) and is_regressor(model_b))
+    ):
+        raise ValueError("Both models must be classifiers or regressors.")
+
     evaluator_a = ModelEvaluator(model_a)
     evaluator_b = ModelEvaluator(model_b)
 
     try:
-        scores_a = evaluator_a.cross_validation_scores(X, y, cv=cv, scoring=scoring)
-        scores_b = evaluator_b.cross_validation_scores(X, y, cv=cv, scoring=scoring)
+        if isinstance(cv, int):
+            if is_classifier(model_a):
+                cv_splitter = StratifiedKFold(n_splits=cv)
+            else:
+                cv_splitter = KFold(n_splits=cv)
+        else:
+            cv_splitter = cv
+
+        scores_a = evaluator_a.cross_validation_scores(
+            X, y, cv=cv_splitter, scoring=scoring
+        )
+        scores_b = evaluator_b.cross_validation_scores(
+            X, y, cv=cv_splitter, scoring=scoring
+        )
 
         results = {}
-        for metric in scores_a.keys():
+        for metric in scores_a:
             mean_a = np.mean(scores_a[metric])
             mean_b = np.mean(scores_b[metric])
             _, p_value = ttest_rel(scores_a[metric], scores_b[metric])
@@ -205,5 +218,5 @@ def perform_ab_test(
 
         return results
     except Exception as e:
-        logger.error(f"Error occurred during A/B testing: {e}")
-        raise RuntimeError(f"A/B testing failed: {e}")
+        logger.error(f"A/B testing error: {e}")
+        raise RuntimeError("A/B testing failed.") from e
